@@ -1,13 +1,17 @@
+"""Package initialization."""
+
 import datetime
 import json
 
 import pkg_resources
 
+from . import dbms, dir, geom, ops, settings, store, text
+
 with open(pkg_resources.resource_filename(__name__, "data/metadata.json"), mode='r') as metadata_file:
     metadata = json.load(metadata_file)
 
 __project__ = metadata['Project']
-__package__ = metadata['Package']
+__pkgname__ = metadata['Package']
 
 __author__ = metadata['Author']
 __affiliation__ = metadata['Affiliation']
@@ -19,3 +23,13 @@ __copyright__ = f'2019-{datetime.datetime.now().year}, {__author__}'
 
 __version__ = metadata['Version']
 __license__ = metadata['License']
+
+__all__ = [
+    'dbms',
+    'dir',
+    'geom',
+    'ops',
+    'settings',
+    'store',
+    'text',
+]
