@@ -1,12 +1,12 @@
-"""
-Miscellaneous operations.
-"""
+"""Miscellaneous operations."""
 
 import ast
 import collections.abc
 import copy
 import datetime
+import hashlib
 import html.parser
+import importlib
 import inspect
 import itertools
 import json
@@ -14,6 +14,7 @@ import math
 import os
 import random
 import re
+import secrets
 import shutil
 import socket
 import sys
@@ -25,12 +26,14 @@ import pandas as pd
 import pkg_resources
 import requests
 import requests.adapters
-import scipy.sparse
 import urllib3.util.retry
 
-import pyhelpers._cache
+from ._cache import _check_dependency, _USER_AGENT_STRINGS
 
-""" == General use =========================================================================== """
+
+# ==================================================================================================
+# General use
+# ==================================================================================================
 
 
 def confirmed(prompt=None, confirmation_required=True, resp=False):
@@ -48,7 +51,7 @@ def confirmed(prompt=None, confirmation_required=True, resp=False):
     :return: a response
     :rtype: bool
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import confirmed
 
@@ -82,7 +85,7 @@ def confirmed(prompt=None, confirmation_required=True, resp=False):
         return True
 
 
-def get_obj_attr(obj, col_names=None):
+def get_obj_attr(obj, col_names=None, as_dataframe=False):
     """
     Get main attributes of an object.
 
@@ -90,7 +93,9 @@ def get_obj_attr(obj, col_names=None):
     :type obj: object
     :param col_names: a list of column names
     :type col_names: list
-    :return: tabular data of the main attributes of the given object
+    :param as_dataframe: whether to return the data in tabular format, defaults to ``False``
+    :type as_dataframe: bool
+    :return: list or tabular data of the main attributes of the given object
     :rtype: pandas.DataFrame
 
     **Examples**::
@@ -102,9 +107,9 @@ def get_obj_attr(obj, col_names=None):
         Password (postgres@localhost:5432): ***
         Connecting postgres:***@localhost:5432/postgres ... Successfully.
 
-        >>> obj_attr = get_obj_attr(postgres)
+        >>> obj_attr = get_obj_attr(postgres, as_dataframe=True)
         >>> obj_attr.head()
-                  Attribute       Value
+                  attribute       value
         0  DEFAULT_DATABASE    postgres
         1   DEFAULT_DIALECT  postgresql
         2    DEFAULT_DRIVER    psycopg2
@@ -130,15 +135,16 @@ def get_obj_attr(obj, col_names=None):
     """
 
     if col_names is None:
-        col_names = ['Attribute', 'Value']
+        col_names = ['attribute', 'value']
 
     all_attrs = inspect.getmembers(obj, lambda x: not (inspect.isroutine(x)))
 
     attrs = [x for x in all_attrs if not re.match(r'^__?', x[0])]
 
-    attrs_tbl = pd.DataFrame(attrs, columns=col_names)
+    if as_dataframe:
+        attrs = pd.DataFrame(attrs, columns=col_names)
 
-    return attrs_tbl
+    return attrs
 
 
 def eval_dtype(str_val):
@@ -182,7 +188,7 @@ def gps_to_utc(gps_time):
     :return: UTC time
     :rtype: datetime.datetime
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import gps_to_utc
 
@@ -318,9 +324,9 @@ def find_executable(app_name, possibilities=None):
     :param app_name: executable filename of the application that is to be called
     :type app_name: str
     :param possibilities: possible pathnames
-    :type possibilities: list or None
-    :return: pathname of the executable file
-    :rtype: str
+    :type possibilities: list or set or None
+    :return: pathname of the specified executable file and whether it exists
+    :rtype: typing.Tuple[str, bool]
 
     **Examples**::
 
@@ -330,33 +336,145 @@ def find_executable(app_name, possibilities=None):
         >>> python_exe = "python.exe"
         >>> possible_paths = ["C:\\Program Files\\Python39", "C:\\Python39"]
 
-        >>> path_to_python_exe = find_executable(app_name=python_exe, possibilities=possible_paths)
+        >>> path_to_python_exe, python_exe_exists = find_executable(python_exe, possible_paths)
         >>> os.path.relpath(path_to_python_exe)
         'venv\\Scripts\\python.exe'
+        >>> python_exe_exists
+        True
 
         >>> text_exe = "pyhelpers.exe"  # This file does not actually exist
-        >>> path_to_test_exe = find_executable(app_name=text_exe, possibilities=possible_paths)
+        >>> path_to_test_exe, test_exe_exists = find_executable(text_exe, possible_paths)
         >>> path_to_test_exe
         'pyhelpers.exe'
+        >>> test_exe_exists
+        False
     """
 
-    exe = copy.copy(app_name)
+    exe_pathname = copy.copy(app_name)
+    exe_exists = False
 
-    if not os.path.isfile(exe):
-        alt_exe_pathnames = [shutil.which(app_name)]
+    if not os.path.isfile(exe_pathname):
+        alt_exe_pathnames = {shutil.which(app_name)}
         if possibilities is not None:
-            alt_exe_pathnames += possibilities
+            alt_exe_pathnames.update(set(possibilities))
 
-        for exe_pathname in alt_exe_pathnames:
-            if exe_pathname:
-                if os.path.isfile(exe_pathname):
-                    exe = exe_pathname
+        for x in alt_exe_pathnames:
+            if x:
+                if os.path.isfile(x):
+                    exe_pathname = x
+                    exe_exists = True
                     break
 
-    return exe
+    return exe_exists, exe_pathname
 
 
-""" == Basic data manipulation =============================================================== """
+def hash_password(password, salt=None, salt_size=None, iterations=None, ret_hash=True, **kwargs):
+    """
+    Hash a password using `hashlib.pbkdf2_hmac
+    <https://docs.python.org/3/library/hashlib.html#hashlib.pbkdf2_hmac>`_.
+
+    See also [`OPS-HP-1 <https://nitratine.net/blog/post/how-to-hash-passwords-in-python/>`_].
+
+    :param password: input as a password
+    :type password: str or int or float
+    :param salt: random data; when ``salt=None`` (default), it is generated by `os.urandom()`_,
+        which depends on ``salt_size``;
+        see also [`OPS-HP-2 <https://en.wikipedia.org/wiki/Salt_%28cryptography%29>`_]
+    :type salt: bytes or str
+    :param salt_size: ``size`` of the function `os.urandom()`_, i.e. the size of a random bytestring
+        for cryptographic use; when ``salt_size=None`` (default), it uses ``128``
+    :type salt_size: int or None
+    :param iterations: ``size`` of the function `hashlib.pbkdf2_hmac()`_,
+        i.e. number of iterations of SHA-256; when ``salt_size=None`` (default), it uses ``100000``
+    :type iterations: int or None
+    :param ret_hash: whether to return the salt and key, defaults to ``True``
+    :type ret_hash: bool
+    :param kwargs: [optional] parameters of the function `hashlib.pbkdf2_hmac()`_
+    :return: (only when ``ret_hash=True``) salt and key
+    :rtype: bytes
+
+    .. _`os.urandom()`: https://docs.python.org/3/library/os.html#os.urandom
+    .. _`hashlib.pbkdf2_hmac()`: https://docs.python.org/3/library/hashlib.html#hashlib.pbkdf2_hmac
+
+    **Examples**::
+
+        >>> from pyhelpers.ops import hash_password, verify_password
+
+        >>> sk = hash_password('test%123', salt_size=16)  # salt and key
+
+        >>> salt_data = sk[:16].hex()
+        >>> key_data = sk[16:].hex()
+
+        >>> verify_password('test%123', salt=salt_data, key=key_data)
+        True
+    """
+
+    if not isinstance(password, (bytes, bytearray)):
+        pwd = str(password).encode('UTF-8')
+    else:
+        pwd = password
+
+    if salt is None:
+        salt_ = os.urandom(128 if salt_size is None else salt_size)
+    else:
+        salt_ = salt.encode() if isinstance(salt, str) else salt
+
+    iterations_ = 100000 if iterations is None else iterations
+
+    key = hashlib.pbkdf2_hmac(
+        hash_name='SHA256', password=pwd, salt=salt_, iterations=iterations_, **kwargs)
+
+    if ret_hash:
+        salt_and_key = salt_ + key
+        return salt_and_key
+
+
+def verify_password(password, salt, key, iterations=None):
+    """
+    Verify a password given salt and key.
+
+    :param password: input as a password
+    :type password: str or int or float
+    :param salt: random data;
+        see also [`OPS-HP-1 <https://en.wikipedia.org/wiki/Salt_%28cryptography%29>`_]
+    :type salt: bytes or str
+    :param key: PKCS#5 password-based key (produced by the function `hashlib.pbkdf2_hmac()`_)
+    :type key: bytes or str
+    :param iterations: ``size`` of the function `hashlib.pbkdf2_hmac()`_,
+        i.e. number of iterations of SHA-256; when ``salt_size=None`` (default), it uses ``100000``
+    :type iterations: int or None
+    :return: whether the input password is correct
+    :rtype: bool
+
+    .. _`hashlib.pbkdf2_hmac()`: https://docs.python.org/3/library/hashlib.html#hashlib.pbkdf2_hmac
+
+    .. seealso::
+
+        - Examples of the function :py:func:`pyhelpers.ops.hash_password`
+    """
+
+    pwd = str(password).encode('UTF-8') if not isinstance(password, (bytes, bytearray)) else password
+    iterations_ = 100000 if iterations is None else iterations
+
+    def _is_hex(x):
+        try:
+            int(x, 16)
+            return True
+        except ValueError:
+            return False
+
+    key_ = hashlib.pbkdf2_hmac(
+        hash_name='SHA256', password=pwd, salt=bytes.fromhex(salt) if _is_hex(salt) else salt,
+        iterations=iterations_)
+
+    rslt = True if key_ == (bytes.fromhex(key) if _is_hex(key) else key) else False
+
+    return rslt
+
+
+# ==================================================================================================
+# Basic data manipulation
+# ==================================================================================================
 
 
 # Iterable
@@ -402,7 +520,7 @@ def split_list_by_size(lst, sub_len):
     :return: a sequence of ``sub_len``-sized sub-lists from ``lst``
     :rtype: typing.Generator[list]
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import split_list_by_size
 
@@ -432,7 +550,7 @@ def split_list(lst, num_of_sub):
     :return: a total of ``num_of_sub`` sub-lists from ``lst``
     :rtype: typing.Generator[list]
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import split_list
 
@@ -728,7 +846,7 @@ def remove_dict_keys(dictionary, *keys):
     :param keys: (a sequence of) any that can be the key of a dictionary
     :type keys: any
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import remove_dict_keys
 
@@ -855,7 +973,7 @@ def detect_nan_for_str_column(data_frame, column_names=None):
     :return: position index of the column that contains ``NaN``
     :rtype: typing.Generator[typing.Iterable]
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import detect_nan_for_str_column
         >>> from pyhelpers._cache import example_dataframe
@@ -863,14 +981,16 @@ def detect_nan_for_str_column(data_frame, column_names=None):
         >>> dat = example_dataframe()
         >>> dat
                     Easting  Northing
+        City
         London       530034    180381
         Birmingham   406689    286822
         Manchester   383819    398052
         Leeds        582044    152953
 
-        >>> dat.iloc[3, 1] = None
+        >>> dat.loc['Leeds', 'Northing'] = None
         >>> dat
                     Easting  Northing
+        City
         London       530034  180381.0
         Birmingham   406689  286822.0
         Manchester   383819  398052.0
@@ -899,7 +1019,7 @@ def create_rotation_matrix(theta):
     :return: a rotation matrix of shape (2, 2)
     :rtype: numpy.ndarray
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import create_rotation_matrix
 
@@ -929,7 +1049,7 @@ def dict_to_dataframe(input_dict, k='key', v='value'):
     :return: a data frame converted from the ``input_dict``
     :rtype: pandas.DataFrame
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import dict_to_dataframe
 
@@ -965,7 +1085,7 @@ def parse_csr_matrix(path_to_csr, verbose=False, **kwargs):
 
     .. _`numpy.load`: https://numpy.org/doc/stable/reference/generated/numpy.load
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import parse_csr_matrix
         >>> from pyhelpers.dir import cd
@@ -997,6 +1117,8 @@ def parse_csr_matrix(path_to_csr, verbose=False, **kwargs):
         True
     """
 
+    scipy_sparse = _check_dependency(name='scipy.sparse')
+
     if verbose:
         print("Loading \"\\{}\"".format(os.path.relpath(path_to_csr)), end=" ... ")
 
@@ -1007,9 +1129,10 @@ def parse_csr_matrix(path_to_csr, verbose=False, **kwargs):
         indptr = csr_loader['indptr']
         shape = csr_loader['shape']
 
-        csr_mat = scipy.sparse.csr_matrix((data, indices, indptr), shape)
+        csr_mat = scipy_sparse.csr_matrix((data, indices, indptr), shape)
 
-        print("Done.") if verbose else ""
+        if verbose:
+            print("Done.")
 
         return csr_mat
 
@@ -1115,7 +1238,65 @@ def swap_rows(array, r1, r2, as_list=False):
     return array_
 
 
-""" == Basic computation ===================================================================== """
+def np_shift(array, step, fill_value=np.nan):
+    """
+    Shift an array by desired number of rows.
+
+    See also [`OPS-NS-1 <https://stackoverflow.com/questions/30399534/>`_]
+
+    :param array: an array of numbers
+    :type array: numpy.ndarray
+    :param step: number of rows to shift
+    :type step: int
+    :param fill_value: values to fill missing rows due to the shift, defaults to ``NaN``
+    :type fill_value: float or int
+    :return: shifted array
+    :rtype: numpy.ndarray
+
+    **Examples**::
+
+        >>> from pyhelpers.ops import np_shift
+        >>> from pyhelpers._cache import example_dataframe
+
+        >>> arr = example_dataframe().to_numpy()
+        >>> arr
+        array([[530034, 180381],
+               [406689, 286822],
+               [383819, 398052],
+               [582044, 152953]], dtype=int64)
+
+        >>> np_shift(arr, step=-1)
+        array([[406689., 286822.],
+               [383819., 398052.],
+               [582044., 152953.],
+               [    nan,     nan]])
+
+        >>> np_shift(arr, step=1, fill_value=0)
+        array([[     0,      0],
+               [530034, 180381],
+               [406689, 286822],
+               [383819, 398052]], dtype=int64)
+    """
+
+    result = np.empty_like(array, dtype=type(fill_value))  # np.zeros_like(array)
+
+    if step > 0:
+        result[:step] = fill_value
+        result[step:] = array[:-step]
+
+    elif step < 0:
+        result[step:] = fill_value
+        result[:step] = array[-step:]
+
+    else:
+        result[:] = array
+
+    return result
+
+
+# ==================================================================================================
+# Basic computation
+# ==================================================================================================
 
 
 def get_extreme_outlier_bounds(num_dat, k=1.5):
@@ -1129,7 +1310,7 @@ def get_extreme_outlier_bounds(num_dat, k=1.5):
     :return: lower and upper bound
     :rtype: tuple
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import get_extreme_outlier_bounds
         >>> import pandas
@@ -1188,7 +1369,7 @@ def interquartile_range(num_dat):
     :return: interquartile range of ``num_dat``
     :rtype: float
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import interquartile_range
 
@@ -1259,7 +1440,9 @@ def find_closest_date(date, lookup_dates, as_datetime=False, fmt='%Y-%m-%d %H:%M
     return closest_date
 
 
-""" == Graph plotting ======================================================================== """
+# ==================================================================================================
+# Graph plotting
+# ==================================================================================================
 
 
 def cmap_discretisation(cmap, n_colours):
@@ -1280,7 +1463,7 @@ def cmap_discretisation(cmap, n_colours):
     .. _`colormaps`: https://matplotlib.org/tutorials/colors/colormaps.html
     .. _`matplotlib.cm.get_cmap`: https://matplotlib.org/api/cm_api.html#matplotlib.cm.get_cmap
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import cmap_discretisation
         >>> import matplotlib.cm
@@ -1299,25 +1482,24 @@ def cmap_discretisation(cmap, n_colours):
         >>> plt.tight_layout()
         >>> plt.show()
 
-    The exmaple is illustrated in :numref:`cmap-discretisation`:
+    The exmaple is illustrated in :numref:`ops-cmap_discretisation-demo`:
 
-    .. figure:: ../_images/cmap-discretisation.*
-        :name: cmap-discretisation
+    .. figure:: ../_images/ops-cmap_discretisation-demo.*
+        :name: ops-cmap_discretisation-demo
         :align: center
         :width: 60%
 
         An example of discrete colour ramp, created by the function
-        :py:func:`cmap_discretisation()<pyhelpers.ops.cmap_discretisation>`.
+        :py:func:`~pyhelpers.ops.cmap_discretisation`.
 
     .. code-block:: python
 
         >>> plt.close()
     """
 
-    import matplotlib.cm
-    import matplotlib.colors
+    matplotlib_cm, matplotlib_colors = map(_check_dependency, ['matplotlib.cm', 'matplotlib.colors'])
 
-    cmap_ = matplotlib.cm.get_cmap(cmap) if isinstance(cmap, str) else copy.copy(cmap)
+    cmap_ = matplotlib_cm.get_cmap(cmap) if isinstance(cmap, str) else copy.copy(cmap)
 
     colours_i = np.concatenate((np.linspace(0, 1., n_colours), (0., 0., 0., 0.)))
     colours_rgba = cmap_(colours_i)
@@ -1329,7 +1511,7 @@ def cmap_discretisation(cmap, n_colours):
             (indices[x], colours_rgba[x - 1, ki], colours_rgba[x, ki]) for x in range(n_colours + 1)
         ]
 
-    colour_map = matplotlib.colors.LinearSegmentedColormap(cmap.name + '_%d' % n_colours, c_dict, 1024)
+    colour_map = matplotlib_colors.LinearSegmentedColormap(cmap.name + '_%d' % n_colours, c_dict, 1024)
 
     return colour_map
 
@@ -1377,15 +1559,15 @@ def colour_bar_index(cmap, n_colours, labels=None, **kwargs):
         >>> plt.tight_layout()
         >>> plt.show()
 
-    The above example is illustrated in :numref:`colour-bar-index-1`:
+    The above example is illustrated in :numref:`ops-colour_bar_index-demo-1`:
 
-    .. figure:: ../_images/colour-bar-index-1.*
-        :name: colour-bar-index-1
+    .. figure:: ../_images/ops-colour_bar_index-demo-1.*
+        :name: ops-colour_bar_index-demo-1
         :align: center
         :width: 23%
 
         An example of colour bar with numerical index,
-        created by the function :py:func:`colour_bar_index()<pyhelpers.ops.colour_bar_index>`.
+        created by the function :py:func:`~pyhelpers.ops.colour_bar_index`.
 
     .. code-block:: python
 
@@ -1402,31 +1584,30 @@ def colour_bar_index(cmap, n_colours, labels=None, **kwargs):
         >>> plt.tight_layout()
         >>> plt.show()
 
-    This second example is illustrated in :numref:`colour-bar-index-2`:
+    This second example is illustrated in :numref:`ops-colour_bar_index-demo-2`:
 
-    .. figure:: ../_images/colour-bar-index-2.*
-        :name: colour-bar-index-2
+    .. figure:: ../_images/ops-colour_bar_index-demo-2.*
+        :name: ops-colour_bar_index-demo-2
         :align: center
         :width: 23%
 
         An example of colour bar with textual index,
-        created by the function :py:func:`colour_bar_index()<pyhelpers.ops.colour_bar_index>`.
+        created by the function :py:func:`~pyhelpers.ops.colour_bar_index`.
 
     .. code-block:: python
 
         >>> plt.close(fig='all')
     """
 
-    import matplotlib.cm
-    import matplotlib.pyplot
+    matplotlib_cm, matplotlib_pyplot = map(_check_dependency, ['matplotlib.cm', 'matplotlib.pyplot'])
 
     cmap = cmap_discretisation(cmap, n_colours)
 
-    mappable = matplotlib.cm.ScalarMappable(cmap=cmap)
+    mappable = matplotlib_cm.ScalarMappable(cmap=cmap)
     mappable.set_array(np.array([]))
     mappable.set_clim(-0.5, n_colours + 0.5)
 
-    colour_bar = matplotlib.pyplot.colorbar(mappable, **kwargs)
+    colour_bar = matplotlib_pyplot.colorbar(mappable, **kwargs)
     colour_bar.set_ticks(np.linspace(0, n_colours, n_colours))
     colour_bar.set_ticklabels(range(n_colours))
 
@@ -1436,7 +1617,9 @@ def colour_bar_index(cmap, n_colours, labels=None, **kwargs):
     return colour_bar
 
 
-""" == Web data extraction =================================================================== """
+# ==================================================================================================
+# Web data extraction
+# ==================================================================================================
 
 
 def is_network_connected():
@@ -1457,7 +1640,12 @@ def is_network_connected():
     host_name = socket.gethostname()
     ip_address = socket.gethostbyname(host_name)
 
-    return False if ip_address == "127.0.0.1" else True
+    if ip_address == '127.0.0.1':
+        connected = False
+    else:
+        connected = True
+
+    return connected
 
 
 def is_url(url, partially=False):
@@ -1565,7 +1753,7 @@ def is_downloadable(url, request_field='content-type', **kwargs):
 
     .. _`requests.head`: https://2.python-requests.org/en/master/api/#requests.head
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import is_downloadable
 
@@ -1614,7 +1802,7 @@ def init_requests_session(url, max_retries=5, backoff_factor=0.1, retry_status='
     .. _`requests.Session`:
         https://2.python-requests.org/en/master/api/#request-sessions
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import init_requests_session
 
@@ -1643,6 +1831,41 @@ def init_requests_session(url, max_retries=5, backoff_factor=0.1, retry_status='
     return session
 
 
+class _FakeUserAgentParser(html.parser.HTMLParser):
+    def __init__(self):
+        super().__init__()
+        self.reset()
+        self.recording = 0
+        self.data = []
+
+    def error(self, message):
+        pass
+
+    def handle_starttag(self, tag, attrs):
+        if tag != 'a':
+            return
+
+        if self.recording:
+            self.recording += 1
+            return
+
+        if tag == 'a':
+            for name, link in attrs:
+                if name == 'href' and link.startswith('/index.php?id='):
+                    break
+                else:
+                    return
+            self.recording = 1
+
+    def handle_endtag(self, tag):
+        if tag == 'a' and self.recording:
+            self.recording -= 1
+
+    def handle_data(self, data):
+        if self.recording:
+            self.data.append(data.strip())
+
+
 def _user_agent_strings(browser_names=None, dump_dat=True):
     """
     Get a dictionary of user-agent strings for popular browsers.
@@ -1652,7 +1875,7 @@ def _user_agent_strings(browser_names=None, dump_dat=True):
     :return: a dictionary of user-agent strings for popular browsers
     :rtype: dict
 
-    **Example**::
+    **Examples**::
 
         from pyhelpers.ops import _user_agent_strings
 
@@ -1661,40 +1884,6 @@ def _user_agent_strings(browser_names=None, dump_dat=True):
         # ['Chrome', 'Firefox', 'Safari', 'Edge', 'Internet Explorer', 'Opera']
     """
 
-    class _FakeUserAgentParser(html.parser.HTMLParser):
-        def __init__(self):
-            super().__init__()
-            self.reset()
-            self.recording = 0
-            self.data = []
-
-        def error(self, message):
-            pass
-
-        def handle_starttag(self, tag, attrs):
-            if tag != 'a':
-                return
-
-            if self.recording:
-                self.recording += 1
-                return
-
-            if tag == 'a':
-                for name, link in attrs:
-                    if name == 'href' and link.startswith('/index.php?id='):
-                        break
-                    else:
-                        return
-                self.recording = 1
-
-        def handle_endtag(self, tag):
-            if tag == 'a' and self.recording:
-                self.recording -= 1
-
-        def handle_data(self, data):
-            if self.recording:
-                self.data.append(data.strip())
-
     if browser_names is None:
         browser_names_ = ['Chrome', 'Firefox', 'Safari', 'Edge', 'Internet Explorer', 'Opera']
     else:
@@ -1704,13 +1893,9 @@ def _user_agent_strings(browser_names=None, dump_dat=True):
 
     user_agent_strings = {}
     for browser_name in browser_names_:
-        url = resource_url + f'?name={browser_name.replace(" ", "+")}'
-        response = urllib.request.urlopen(url=url)
-        text = response.read().decode('utf-8')
-
+        response = requests.get(url=resource_url + f'?name={browser_name.replace(" ", "+")}')
         fua_parser = _FakeUserAgentParser()
-        fua_parser.feed(text)
-
+        fua_parser.feed(response.text)
         user_agent_strings[browser_name] = list(set(fua_parser.data))
 
     if dump_dat:
@@ -1777,7 +1962,7 @@ def load_user_agent_strings(shuffled=False, flattened=False, update=False, verbo
         # path_to_json = pkg_resources.resource_filename(__name__, "data\\user-agent-strings.json")
         # json_in = open(path_to_json, mode='r')
         # user_agent_strings = json.loads(json_in.read())
-        user_agent_strings = pyhelpers._cache._USER_AGENT_STRINGS
+        user_agent_strings = _USER_AGENT_STRINGS.copy()
 
     else:
         if verbose:
@@ -1785,7 +1970,9 @@ def load_user_agent_strings(shuffled=False, flattened=False, update=False, verbo
 
         try:
             user_agent_strings = _user_agent_strings(dump_dat=True)
-            pyhelpers._cache._USER_AGENT_STRINGS = user_agent_strings
+
+            importlib.reload(sys.modules.get('pyhelpers._cache'))
+
             if verbose:
                 print("Done.")
 
@@ -1852,7 +2039,7 @@ def get_user_agent_string(fancy=None, **kwargs):
         kwargs.update({'flattened': True})
         user_agent_strings = load_user_agent_strings(**kwargs)
 
-    user_agent_string = random.choice(user_agent_strings)
+    user_agent_string = secrets.choice(user_agent_strings)
 
     return user_agent_string
 
@@ -1902,7 +2089,16 @@ def fake_requests_headers(randomized=True, **kwargs):
 
 
 def _download_file_from_url(response, path_to_file):
-    import tqdm
+    """
+    Download an object from a valid URL (and save it as a file).
+
+    :param response: a server's response to an HTTP request
+    :type response: requests.Response
+    :param path_to_file: a path where the downloaded object is saved as, or a filename
+    :type path_to_file: str or os.PathLike[str]
+    """
+
+    tqdm_ = _check_dependency(name='tqdm')
 
     file_size = int(response.headers.get('content-length'))  # Total size in bytes
 
@@ -1912,9 +2108,9 @@ def _download_file_from_url(response, path_to_file):
 
     total_iter = file_size // chunk_size
 
-    progress = tqdm.tqdm(
-        desc=f'"{os.path.relpath(path_to_file)}"', total=total_iter, unit='B',
-        unit_scale=True, unit_divisor=unit_divisor)
+    progress = tqdm_.tqdm(
+        desc=f'"{os.path.relpath(path_to_file)}"', total=total_iter, unit='B', unit_scale=True,
+        unit_divisor=unit_divisor)
 
     contents = response.iter_content(chunk_size=chunk_size, decode_unicode=True)
     with open(file=path_to_file, mode='wb') as f:
@@ -1950,7 +2146,7 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
     :param url: valid URL to a web resource
     :type url: str
     :param path_to_file: a path where the downloaded object is saved as, or a filename
-    :type path_to_file: str
+    :type path_to_file: str or os.PathLike[str]
     :param if_exists: given that the specified file already exists, options include
         ``'replace'`` (default, continuing to download the requested file and replace the existing one
         at the specified path) and ``'pass'`` (cancelling the download)
@@ -1972,7 +2168,7 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
     .. _`requests.Session.get()`:
         https://docs.python-requests.org/en/master/_modules/requests/sessions/#Session.get
 
-    **Example**::
+    **Examples**::
 
         >>> from pyhelpers.ops import download_file_from_url
         >>> from pyhelpers.dir import cd
@@ -1980,7 +2176,7 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
         >>> import os
 
         >>> logo_url = 'https://www.python.org/static/community_logos/python-logo-master-v3-TM.png'
-        >>> path_to_img = cd("tests", "images", "python-logo.png")
+        >>> path_to_img = cd("tests", "images", "ops-download_file_from_url-demo.png")
 
         >>> # Check if "python-logo.png" exists at the specified path
         >>> os.path.exists(path_to_img)
@@ -1996,8 +2192,8 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
         >>> img = Image.open(path_to_img)
         >>> img.show()  # as illustrated below
 
-    .. figure:: ../_images/python-logo.*
-        :name: python-logo
+    .. figure:: ../_images/ops-download_file_from_url-demo.*
+        :name: ops-download_file_from_url-demo
         :align: center
         :width: 65%
 
@@ -2009,12 +2205,9 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
 
     .. note::
 
-        - When setting ``verbose`` to be ``True`` (or ``1``), the function relies on `tqdm`_,
-          which is not an essential dependency for installing `pyhelpers>=1.2.15`_.
-          You may need to install `tqdm`_ before proceeding with ``verbose=True`` (or ``verbose=1``).
+        - When ``verbose=True``, the function requires `tqdm`_.
 
         .. _`tqdm`: https://pypi.org/project/tqdm/
-        .. _`pyhelpers>=1.2.15`: https://pypi.org/project/pyhelpers/1.2.15/
     """
 
     path_to_dir = os.path.dirname(path_to_file)
@@ -2030,6 +2223,9 @@ def download_file_from_url(url, path_to_file, if_exists='replace', max_retries=5
                   f"The download is cancelled.")
 
     else:
+        # if verbose:
+        #     _ = _check_dependency(name='tqdm')
+
         if requests_session_args is None:
             requests_session_args = {}
         session = init_requests_session(url=url, max_retries=max_retries, **requests_session_args)
